package moreswing.swing

import event. {

  InternalFrameClosing,
  InternalFrameMaximized,
  InternalFrameShifted

}

import java.awt.event. { MouseListener, MouseMotionListener }

import javax.swing.plaf.basic.BasicInternalFrameUI

import scala.collection.mutable. { Buffer, Map }

import scala.swing.Rectangle

import scala.swing.event. {

  MouseEntered,
  MousePressed,
  MouseReleased,
  UIElementResized

}

/** Desktop where all internal frames can be in a grid.
 *
 * @author myst3r10n
 */
class GridDesktopPane extends DesktopPane with GridBoundaries {

  // Event handling.
  listenTo(this, mouse.clicks, mouse.moves)
  reactions += {

    case InternalFrameClosing(iframe: InternalFrame) =>
      // Disable arrange by resize.
      arrangeByResize = ""

      // Index of internal frame.
      val n = iframeBuffer.indexOf(iframe)

      // Disable desktop mode gridable.
      gridableDesktop = false

      // Removes internal frame.
      iframeBuffer.remove(n)

      // Removes bounds of internal frame.
      iframeBoundaries.remove(iframe)

    case InternalFrameMaximized(iframe: InternalFrame) =>
      // Disable arrange by resize.
      arrangeByResize = ""

      // Disable desktop mode gridable.
      gridableDesktop = false

    case UIElementResized(desktop: DesktopPane) =>
      // Arrange by resize.
      arrangeByResize match {

	case "horizontal" => horizontal
	case "tile" => tile
	case "vertical" => vertical

	// Disable desktop mode gridable.
	case _ => gridableDesktop = false

      }

    case UIElementResized(iframe: InternalFrame) =>
      // Pass if bounds exists in table.
      if(iframeBoundaries.contains(iframe)) {

	// Pass if maximum.
	if(iframe.maximum)

	  // Removes bounds from table.
	  iframeBoundaries.remove(iframe)

	// Pass if not maximum and gridable desktop enabled.
	else if(gridableDesktop) {

	  // Pass if lastest size exists and unequal internal frame size.
	  if(iframeBoundaries(iframe) != iframe.bounds) {

	    // Disable arrange by resize.
	    arrangeByResize = ""

	    // Deny some boundary motion if internal frame beside global bounds.
	    if((iframeBoundaries(iframe).x < iframe.bounds.x &&
		iframeBoundaries(iframe).x == 0) ||
	       (iframeBoundaries(iframe).x == iframe.bounds.x &&
		iframeBoundaries(iframe).width > iframe.bounds.width &&
		iframeBoundaries(iframe).x + iframeBoundaries(iframe).width == size.width) ||
	       (iframeBoundaries(iframe).y < iframe.bounds.y &&
		iframeBoundaries(iframe).y == 0) ||
	       (iframeBoundaries(iframe).y == iframe.bounds.y &&
		iframeBoundaries(iframe).height > iframe.bounds.height &&
		iframeBoundaries(iframe).y + iframeBoundaries(iframe).height == size.height))
	      iframe.bounds = iframeBoundaries(iframe)

	    // Fit internal frames recursively.
	    else
	      fitBoundaries(iframeBuffer.indexOf(iframe), iframe.bounds,
		  (for(i <- 0 to iframeBoundaries.size - 1) yield(i -> iframeBoundaries(iframeBuffer(i)).asInstanceOf[Rectangle])).toMap,
		  (for(i <- 0 to iframeBuffer.size - 1) yield(i -> iframeBuffer(i).minimumSize)).toMap).foreach { case (modified, modifiedBounds) =>

                if(modifiedBounds != iframeBoundaries(iframeBuffer(modified)))
		  iframeBoundaries += iframeBuffer(modified) -> modifiedBounds
                iframeBuffer(modified).bounds = modifiedBounds

              }

	  // Pass if latest size equal internal frame size.
	  } else

	    // Restores size of internal iframe.
	    iframe.bounds = iframeBoundaries(iframe)

	}
      }
  }

  /** Toggle shift mode flotable. */
  var flotableShift = false

  /** @see [[scala.swing.DesktopPane]]. */
  override def += (iframe: InternalFrame) { insert(iframeBuffer.size, iframe) }

  /** Desktop mode gridable. */
  def gridableDesktop = gridableDesktopMode

  /** Toggle desktop mode gridable.
   *
   * @param b If this desktop mode is gridable.
   */
  def gridableDesktop_= (b: Boolean) {

    // Pass if gridable desktop disable.
    if(b == gridableDesktop)
      return

    // Toggle desktop mode gridable.
    gridableDesktopMode = b

    // Pass if gridable desktop enabled.
    if(gridableDesktop) {

      // Tile internal frames.
      tile

      // Reset former mouse listeners of title bars. */
      titleBarsMouseListeners.clear

      // Reset former mouse motion listeners of title bars. */
      titleBarsMouseMotionListeners.clear

      // Backup newest mouse motion listeners of title bars. */
      iframeBuffer.foreach { iframe =>

	// Get title bar.
	val titleBar = iframe.peer.getUI.asInstanceOf[BasicInternalFrameUI].getNorthPane

        // Pass if title bar available.
        if(titleBar != null) {

	  // Backup newest mouse listeners of title bar. */
          titleBarsMouseListeners += iframe ->titleBar.getListeners(classOf[MouseListener]).asInstanceOf[Array[MouseListener]]

	  // Backup newest mouse motion listeners of title bar. */
          titleBarsMouseMotionListeners += iframe ->titleBar.getListeners(classOf[MouseMotionListener]).asInstanceOf[Array[MouseMotionListener]]

	  // Disable motion desktop. */
          titleBar.getListeners(classOf[MouseMotionListener]).asInstanceOf[Array[MouseMotionListener]].foreach(listener => titleBar.removeMouseMotionListener(listener))

	  // Enabled shifted desktop. */
          iframe.peer.getUI.asInstanceOf[BasicInternalFrameUI].getNorthPane.addMouseListener(new MouseListener {

	    def mousePressed(event: java.awt.event.MouseEvent) {

      	      // Save index of internal frame where mouse hovers.
	      mouseOverIFrame = iframeBuffer.indexOf(iframe)

	      val iframeArriveIn = mouseOverIFrame

	      // Enable shifted desktop.
	      if(iframeDepartFrom < 0 &&
	         iframeArriveIn > -1)
		iframeDepartFrom = iframeArriveIn

	    }

	    def mouseEntered(event: java.awt.event.MouseEvent) {

	      // Save internal frame where mouse hovers.
	      mouseOverIFrame = iframeBuffer.indexOf(iframe)

	      val iframeArriveIn = mouseOverIFrame

	      // Move internal frame if gridable desktop enabled, flotable shift enabled, destination valid and unequal start.
	      if(gridableDesktop &&
		 flotableShift &&
		 iframeDepartFrom > -1 &&
		 iframeArriveIn > -1 &&
		 iframeDepartFrom != iframeArriveIn) {

		shiftBounds(iframeBuffer(iframeDepartFrom), iframeBuffer(iframeArriveIn))
		iframeDepartFrom = iframeArriveIn

	      }
	    }

	    def mouseReleased(event: java.awt.event.MouseEvent) {

	      val iframeArriveIn = mouseOverIFrame

	      // Move internal frame if gridable desktop enabled, flotable shift enabled, destination valid and unequal start.
	      if(gridableDesktop &&
		 !flotableShift &&
		 iframeDepartFrom > -1 &&
		 iframeArriveIn > -1 &&
		 iframeDepartFrom != iframeArriveIn) {

		shiftBounds(iframeBuffer(iframeDepartFrom), iframeBuffer(iframeArriveIn))
		iframeDepartFrom = iframeArriveIn

	      }

	      // Disable motion mode.
	      if(iframeDepartFrom != -1)
		iframeDepartFrom = -1

	    }

            // Ignore.
	    def mouseClicked(event: java.awt.event.MouseEvent) { }
	    def mouseExited(event: java.awt.event.MouseEvent) { }

	  } )
	}
      }

    // Pass if gridable desktop disabled.
    } else

      // Disable shiftable desktop. */
      iframeBuffer.foreach { iframe =>

	// Get title bar.
	val titleBar = iframe.peer.getUI.asInstanceOf[BasicInternalFrameUI].getNorthPane

        // Pass if title bar available.
        if(titleBar != null) {

	  // Disable shiftable desktop. */
          titleBar.getListeners(classOf[MouseListener]).asInstanceOf[Array[MouseListener]].foreach(listener => titleBar.removeMouseListener(listener))

	  // Enable motion desktop. */
          titleBarsMouseListeners(iframe).foreach(listener => titleBar.addMouseListener(listener))
          titleBarsMouseMotionListeners(iframe).foreach(listener => titleBar.addMouseMotionListener(listener))

	}
      }
  }

  /** Arranges the internal frames horizontal. */
  def horizontal {

    val iframes = allFrames
    var i = allFrames.size
    val height = size.height / iframes.size

    // horizontal algorithm.
    while(i > 0) {

      i -= 1

      if(iframes(i).maximum)
        iframes(i).maximum = false

      val bounds = new Rectangle(0, i * height, size.width, height)

      if(gridableDesktop)
        iframeBoundaries += iframes(i) -> bounds

      iframes(i).bounds = bounds

    }

    // Enable arrange by resize.
    arrangeByResize = "horizontal"

/*
    var i = 0
    val height = size.height / iframeBuffer.size

    // horizontal algorithm.
    iframeBuffer.foreach { iframe =>

      if(iframe.maximum)
	iframe.maximum = false

      val bounds = new Rectangle(0, i * height, size.width, height)
      if(gridableDesktop)
	iframeBoundaries += iframe -> bounds
      iframe.bounds = bounds

      i += 1

    }

    // Enable arrange by resize.
    arrangeByResize = "horizontal"
*/
  }

  /** @see [[scala.swing.DesktopPane]]. */
  override def insert(n: Int, iframe: InternalFrame) {

    gridableDesktop = false
    listenTo(iframe, iframe.mouse.clicks, iframe.mouse.moves)
    iframeBuffer.insert(n, iframe)
    peer.add(iframe.peer, n)

  }

  /** Shift bounds of a internal frame.
   *
   * @param from The depart from of internal frame.
   * @param to The arrive in of internal frame.
   */
  def shiftBounds(from: InternalFrame, to: InternalFrame) {

    // Pass if both different location.
    if(from != to) {

      // Save buffer indexes of itnernal frames.
      val fromIndex = iframeBuffer.indexOf(from)
      val toIndex = iframeBuffer.indexOf(to)

      // Remember where internal frame must arrive in.
      val arriveIn = to.bounds

      // Pass if "from" is left from "to" in the buffer.
      if(fromIndex < toIndex) {

	// Moves all bounds between "from" and "to" incl. "to" to right.
	var i = toIndex
	while(i > fromIndex) {

	  // Moves bounds to right.
	  iframeBoundaries += iframeBuffer(i) -> iframeBuffer(i - 1).bounds
	  iframeBuffer(i).bounds = iframeBuffer(i - 1).bounds

	  i -= 1

	}

      // Pass if "to" is right from "from" in the buffer.
      } else if(toIndex < fromIndex) {

	// Moves all bounds between "to" and "from" incl. "from" to left.
	var i = toIndex
	while(i < fromIndex) {

	  // Moves bounds to left.
	  iframeBoundaries += iframeBuffer(i) -> iframeBuffer(i + 1).bounds
	  iframeBuffer(i).bounds = iframeBuffer(i + 1).bounds

	  i += 1

	}
      }

      // Shift "from" bounds to "to".
      iframeBoundaries += from -> arriveIn
      from.bounds = arriveIn

      // Shift "from" to "to".
      iframeBuffer.insert(toIndex, iframeBuffer.remove(fromIndex))

      // Publish shift event.
      publish(InternalFrameShifted(iframeBuffer(fromIndex), iframeBuffer(toIndex)))

    }

  }

  /** Arranges the internal frames tiling. */
  def tile {

    var iframes = allFrames
    var boundaries = Seq[Rectangle]()

    var rows = math.sqrt(iframes.size).toInt
    val cols = iframes.size / rows
    val extra = iframes.size % rows

    val width = size.width / cols
    var height = size.height / rows
    var r = 0
    var c = 0

    // Tile algorithm.
    iframeBuffer.foreach { iframe =>

      boundaries = boundaries :+ new Rectangle(c * width,
					                                     r * height,
					                                     if(c < cols - 1)
					                                       width
					                                     else
					                                       size.width - (c * width),
					                                     if(r < rows - 1)
					                                       height
					                                     else
					                                       size.height - (r * height))

      r += 1

      if(r == rows) {

        r = 0
        c += 1

        if(c == cols - extra) {
          rows += 1
          height = size.height / rows

        }
      }
    }

    // Apply all.
    var i = allFrames.size

    while(i > 0) {

      i -= 1

      if(iframes(i).maximum)
        iframes(i).maximum = false

      if(gridableDesktop)
        iframeBoundaries += iframes(i) -> boundaries(i)
      iframes(i).bounds = boundaries(i)
    }

    // Enable arrange by resize.
    arrangeByResize = "tile"

/*
    var rows = math.sqrt(iframeBuffer.size).toInt
    val cols = iframeBuffer.size / rows
    val extra = iframeBuffer.size % rows

    val width = size.width / cols
    var height = size.height / rows
    var r = 0
    var c = 0

    // Tile algorithm.
    iframeBuffer.foreach { iframe =>

      if(iframe.maximum)
	iframe.maximum = false

      val bounds = new Rectangle(c * width, r * height, if(c < cols - 1) width else size.width - (c * width), if(r < rows - 1) height else size.height - (r * height))
      if(gridableDesktop)
	iframeBoundaries += iframe -> bounds
      iframe.bounds = bounds

      r += 1

      if(r == rows) {

	r = 0
	c += 1

	if(c == cols - extra) {
	  rows += 1
	  height = size.height / rows

	}

      }

    }

    // Enable arrange by resize.
    arrangeByResize = "tile"
*/

  }

  /** Arranges the internal frames vertical. */
  def vertical {

    var iframes = allFrames
    var i = iframes.size
    val width =  size.width / iframeBuffer.size

    // vertical algorithm.
    while(i > 0) {

      i -= 1

      if(iframes(i).maximum)
        iframes(i).maximum = false

      val bounds = new Rectangle(i * width, 0, width , size.height)
      if(gridableDesktop)
        iframeBoundaries += iframes(i) -> bounds
      iframes(i).bounds = bounds

    }

    // Enable arrange by resize.
    arrangeByResize = "vertical"
/*
    var i = 0
    val width =  size.width / iframeBuffer.size

    // vertical algorithm.
    iframeBuffer.foreach { iframe =>

      if(iframe.maximum)
	iframe.maximum = false

      val bounds = new Rectangle(i * width, 0, width , size.height)
      if(gridableDesktop)
	iframeBoundaries += iframe -> bounds
      iframe.bounds = bounds

      i += 1

    }

    // Enable arrange by resize.
    arrangeByResize = "vertical"
*/
  }

  /** Arrange by resize. */
  private var arrangeByResize = ""

  /** Internal frames. */
  protected var iframeBuffer = Buffer[InternalFrame]()

  /** Desktop mode gridable. */
  private var gridableDesktopMode = false

  /** Boundaries of internal frames. */
  private var iframeBoundaries = Map[InternalFrame, Rectangle]()

  /** Flying iframe. */
  private var iframeDepartFrom = -1

  /** IFrame where mouse hover. */
  private var mouseOverIFrame = -1

  /** Backup of mouse listeners of every internal frame Title bar. */
  private var titleBarsMouseListeners = Map[InternalFrame, Array[MouseListener]]()

  /** Backup of mouse motion listeners of every internal frame Title bar. */
  private var titleBarsMouseMotionListeners = Map[InternalFrame, Array[MouseMotionListener]]()

}
